#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 25 09:30:33 2019

@author: Rajiv Sambasivan
"""

from arango import ArangoClient, DatabaseListError
import logging
from arangopipe.arangopipe_storage.arangopipe_config import ArangoPipeConfig
from arangopipe.arangopipe_storage.custom_http_client import CustomHTTPClient
from arangopipe.arangopipe_storage.managed_service_conn_parameters import ManagedServiceConnParam
import json
import requests
#import traceback
import sys
# create logger with 'spam_application'
logger = logging.getLogger('arangopipe_admin_logger')
logger.setLevel(logging.DEBUG)
# create file handler which logs even debug messages
fh = logging.FileHandler('arangopipeadmin.log')
fh.setLevel(logging.DEBUG)
# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.ERROR)
# create formatter and add it to the handlers
formatter = logging.Formatter(
    '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
ch.setFormatter(formatter)
# add the handlers to the logger
logger.addHandler(fh)
logger.addHandler(ch)


class ArangoPipeAdmin:
    def __init__(self, reuse_connection=True, config=None, persist_conn=True):
        self.reuse_connection = reuse_connection
        self.db = None
        self.emlg = None
        self.config = None
        self.cfg = None
        self.mscp = ManagedServiceConnParam()
        self.use_supp_config_to_reconnect = False

        if reuse_connection:
            info_msg = "If a config is provided, it will be used for setting up the connection"
            if config is None:
                self.config = self.create_config()
                self.cfg = self.config.get_cfg()
                self.use_supp_config_to_reconnect = False
            else:
                self.config = config
                self.cfg = config.cfg
                self.use_supp_config_to_reconnect = True

            logger.info(info_msg)
        else:

            assert config is not None,\
                   "You must provide connection information for new connections"

            self.config = config
            self.cfg = config.cfg

        try:

            db_serv_host = self.cfg['arangodb'][self.mscp.DB_SERVICE_HOST]
            db_serv_port = self.cfg['arangodb'][self.mscp.DB_SERVICE_PORT]
            db_end_point = self.cfg['arangodb'][self.mscp.DB_SERVICE_END_POINT]
            db_serv_name = self.cfg['arangodb'][self.mscp.DB_SERVICE_NAME]

        except KeyError as k:
            logger.error("Connection information is missing : " + k.args[0])
            logger.error(
                "Please try again after providing the missing information !")
            raise Exception("Key error associated with missing " + k.args[0])

        # check if connection preferences are indicated
        if 'dbName' in self.cfg['arangodb']:
            logger.info("DB name for connection: " + \
                        str(self.cfg['arangodb'][self.mscp.DB_NAME]))
            db_dbName = self.cfg['arangodb'][self.mscp.DB_NAME]
        else:
            db_dbName = ''
        if 'username' in self.cfg['arangodb']:
            logger.info("user name for connection: " +\
                        str(self.cfg['arangodb'][self.mscp.DB_USER_NAME]))
            db_user_name = self.cfg['arangodb'][self.mscp.DB_USER_NAME]
        else:
            db_user_name = ''
        if 'password' in self.cfg['arangodb']:
            logger.info("A specific password was requested !")
            db_password = self.cfg['arangodb'][self.mscp.DB_PASSWORD]
        else:
            db_password = ''

        if self.mscp.DB_CONN_PROTOCOL in self.cfg['arangodb']:
            db_conn_protocol = self.cfg['arangodb'][self.mscp.DB_CONN_PROTOCOL]
        else:
            db_conn_protocol = "http"

        if self.mscp.DB_REPLICATION_FACTOR in self.cfg['arangodb']:
            db_replication_factor = self.cfg['arangodb'][
                self.mscp.DB_REPLICATION_FACTOR]
        else:
            db_replication_factor = None

        if self.mscp.DB_ROOT_USER in self.cfg['arangodb']:
            logger.info("A root user was specified, persisting...")

        if self.mscp.DB_ROOT_USER_PASSWORD in self.cfg['arangodb']:
            logger.info("A root user password was specified, persisting...")


        self.create_db(db_serv_host, db_serv_port,\
                       db_serv_name, db_end_point,\
                       db_dbName, db_user_name, db_password, db_conn_protocol)

        # If you could create a DB, proceed with provisioning the graph. Otherwise you
        # had an issue creating the database.
        if self.db is not None:
            self.create_enterprise_ml_graph(db_replication_factor)

            if persist_conn:
                self.config.dump_data()

        return

    def check_repeated_creation(self, api_data):
        if not api_data:
            repeated_connection = False
        else:
            try:
                user_name_equal = api_data[self.mscp.DB_USER_NAME] ==\
                self.cfg['arangodb'][self.mscp.DB_USER_NAME]
                password_equal =  api_data[self.mscp.DB_PASSWORD] ==\
                self.cfg['arangodb'][self.mscp.DB_PASSWORD]
                db_name_equal = api_data[self.mscp.DB_NAME] ==\
                self.cfg['arangodb'][self.mscp.DB_NAME]
                repeated_connection = user_name_equal or password_equal or db_name_equal
                if user_name_equal:
                    logger.info(
                        "Attempting to create a connection with an existing username"
                    )
                if db_name_equal:
                    logger.info(
                        "Attempting to create a connection with an existing db name"
                    )
            except KeyError:
                repeated_connection = False

        return repeated_connection

    def set_connection_params(self, config):
        self.cfg = config
        self.cfg.dump_data()
        return

    def create_config(self):
        apc = ArangoPipeConfig()
        return apc

    def get_config(self):
        return self.config

    def create_db(self, db_srv_host, db_srv_port, db_serv_name,\
                  db_end_point, db_dbName, db_user_name, db_password,\
                  db_conn_protocol):

        host_connection = db_conn_protocol + "://" + db_srv_host + ":" + str(
            db_srv_port)
        client = ArangoClient(hosts=host_connection)
        logger.debug("Connection reuse: " + str(self.reuse_connection))
        if not self.reuse_connection:
            API_ENDPOINT = host_connection + "/_db/_system/" + db_end_point + \
                            "/" + db_serv_name
            print("API endpoint: " + API_ENDPOINT)

            if db_dbName:
                logger.info("DB name preferrence: " + str(db_dbName))
            if db_user_name:
                logger.info("DB user name preferrence: " + str(db_user_name))
            if db_password:
                logger.info(
                    "Password preference for managed connection was indicated !"
                )

            api_data = {self.mscp.DB_NAME : db_dbName,\
                        self.mscp.DB_USER_NAME: db_user_name,\
                        self.mscp.DB_PASSWORD: db_password }
            logger.info("Requesting a managed service database...")

            r = requests.post(url=API_ENDPOINT, json=api_data, verify=False)

            if r.status_code == 409 or r.status_code == 400:
                logger.error(
                    "It appears that you are attempting to connecting using \
                             existing connection information. So either set reconnect = True when you create ArangoPipeAdmin or recreate a connection config and try again!"
                )
                return

            assert r.status_code == 200, \
            "Managed DB endpoint is unavailable !, reason: " + r.reason + " err code: " +\
            str(r.status_code)
            result = json.loads(r.text)
            logger.info("Managed service database was created !")
            ms_dbName = result['dbName']
            ms_user_name = result['username']
            ms_password = result['password']
            self.cfg['arangodb'][self.mscp.DB_NAME] = ms_dbName
            self.cfg['arangodb'][self.mscp.DB_USER_NAME] = ms_user_name
            self.cfg['arangodb'][self.mscp.DB_PASSWORD] = ms_password
            self.cfg['arangodb'][self.mscp.DB_SERVICE_HOST] = db_srv_host
            self.cfg['arangodb'][self.mscp.DB_SERVICE_NAME] = db_serv_name
            self.cfg['arangodb'][self.mscp.DB_SERVICE_END_POINT] = db_end_point
            self.cfg['arangodb'][self.mscp.DB_SERVICE_PORT] = db_srv_port
            self.cfg['arangodb'][self.mscp.DB_CONN_PROTOCOL] = db_conn_protocol

        else:
            if self.use_supp_config_to_reconnect:
                ms_dbName = self.cfg['arangodb'][self.mscp.DB_NAME]
                ms_user_name = self.cfg['arangodb'][self.mscp.DB_USER_NAME]
                ms_password = self.cfg['arangodb'][self.mscp.DB_PASSWORD]

            else:
                disk_cfg = ArangoPipeConfig()
                pcfg = disk_cfg.get_cfg()  # persisted config values
                ms_dbName = pcfg['arangodb'][self.mscp.DB_NAME]
                ms_user_name = pcfg['arangodb'][self.mscp.DB_USER_NAME]
                ms_password = pcfg['arangodb'][self.mscp.DB_PASSWORD]
                self.config = disk_cfg
                self.cfg = disk_cfg.cfg
        # Connect to arangopipe database as administrative user.
        #This returns an API wrapper for "test" database.
        print("Host Connection: " + str(host_connection))
        client = ArangoClient(hosts= host_connection,\
                              http_client=CustomHTTPClient(username = ms_user_name,\
                                                           password = ms_password))

        db = client.db(ms_dbName, ms_user_name, ms_password)
        self.db = db

        return

    def create_enterprise_ml_graph(self, db_replication_factor):

        cl = ['project', 'models', 'datasets', 'featuresets', 'modelparams', 'run',\
              'devperf', 'servingperf', 'deployment']

        if self.reuse_connection:
            self.emlg = self.db.graph(self.cfg['mlgraph']['graphname'])
            return

        if not self.db.has_graph(self.cfg['mlgraph']['graphname']):
            self.emlg = self.db.create_graph(self.cfg['mlgraph']['graphname'])
        else:
            self.emlg = self.db.graph(self.cfg['mlgraph']['graphname'])

        for col in cl:
            if not self.emlg.has_vertex_collection(col):
                self.db.create_collection(col, db_replication_factor)
                self.emlg.create_vertex_collection(col)


        from_list = ['project', 'models', 'run', 'run', 'run', 'run',\
                     'deployment', 'deployment', 'deployment', 'deployment',\
                     'featuresets']
        to_list = ['models', 'run', 'modelparams', 'datasets', 'devperf',\
                   'featuresets', 'servingperf', 'models', 'modelparams',\
                   'featuresets', 'datasets']
        edge_names = ['project_models', 'run_models', 'run_modelparams', 'run_datasets',\
                      'run_devperf', 'run_featuresets', 'deployment_servingperf', \
                      'deployment_model', 'deployment_modelparams', 'deployment_featureset',\
                       'featureset_dataset']
        for edge, fromv, tov in zip(edge_names, from_list, to_list):
            if not self.emlg.has_edge_definition(edge):
                self.db.create_collection(edge, edge = True,\
                                          replication_factor = db_replication_factor)
                self.emlg.create_edge_definition(edge_collection = edge,\
                                                      from_vertex_collections = [fromv],\
                                                      to_vertex_collections = [tov] )

        self.cfg['arangodb'][
            self.mscp.DB_REPLICATION_FACTOR] = db_replication_factor

        return

    def register_project(self, p):

        projects = self.emlg.vertex_collection("project")
        proj_reg = projects.insert(p)

        return proj_reg

    def delete_arangomldb(self):

        return

    def register_deployment(self, dep_tag):

        # Execute the query
        cursor = self.db.aql.execute(
            'FOR doc IN run FILTER doc.deployment_tag == @value RETURN doc',
            bind_vars={'value': dep_tag})
        run_docs = [doc for doc in cursor]
        the_run_doc = run_docs[0]
        # Get the model params for the run
        rmpe = self.emlg.edge_collection("run_modelparams")
        edge_dict = rmpe.edges(the_run_doc, direction="out")
        tmp_id = edge_dict["edges"][0]["_to"]
        mpc = self.emlg.edge_collection("modelparams")
        tagged_model_params = mpc.get(tmp_id)
        # Get the model for the run
        rme = self.emlg.edge_collection("run_models")
        edge_dict = rme.edges(the_run_doc, direction="in")
        tm_id = edge_dict["edges"][0]["_from"]
        mc = self.emlg.edge_collection("models")
        tagged_model = mc.get(tm_id)
        # Get the featureset for the run
        rfse = self.emlg.edge_collection("run_featuresets")
        edge_dict = rfse.edges(the_run_doc, direction="out")
        tfid = edge_dict["edges"][0]["_to"]
        tfc = self.emlg.edge_collection("featuresets")
        tagged_featureset = tfc.get(tfid)
        # Create a deployment artifact
        deployment = self.emlg.vertex_collection("deployment")
        deploy_info = {"tag": dep_tag}
        dep_reg = deployment.insert(deploy_info)
        #Link the deployment to the model parameters
        dep_model_params_edge = self.emlg.edge_collection(
            "deployment_modelparams")
        dep_model_params_key = dep_reg["_key"] + "-" + tagged_model_params[
            "_key"]
        the_dep_model_param_edge = { "_key": dep_model_params_key,\
                                "_from": dep_reg["_id"],\
                                "_to": tagged_model_params["_id"]}

        dep_mp_reg = dep_model_params_edge.insert(the_dep_model_param_edge)

        # Link the deployment to the featureset
        dep_featureset_edge = self.emlg.edge_collection(
            "deployment_featureset")
        dep_featureset_key = dep_reg["_key"] + "-" + tagged_featureset["_key"]
        the_dep_featureset_edge = { "_key": dep_featureset_key,\
                                "_from": dep_reg["_id"],\
                                "_to": tagged_featureset["_id"]}
        dep_fs_reg = dep_featureset_edge.insert(the_dep_featureset_edge)

        # Link the deployment to the model
        dep_model_edge = self.emlg.edge_collection("deployment_model")
        dep_featureset_key = dep_reg["_key"] + "-" + tagged_model["_key"]
        the_dep_model_edge = { "_key": dep_featureset_key,\
                                "_from": dep_reg["_id"],\
                                "_to": tagged_model["_id"]}

        dep_model_reg = dep_model_edge.insert(the_dep_model_edge)
        return dep_model_reg

    def add_vertex_to_arangopipe(self, vertex_to_create):
        rf = self.cfg['arangodb'][self.mscp.DB_REPLICATION_FACTOR]

        if not self.db.has_graph(self.cfg['mlgraph']['graphname']):
            self.emlg = self.db.create_graph(self.cfg['mlgraph']['graphname'])
        else:
            self.emlg = self.db.graph(self.cfg['mlgraph']['graphname'])

        #Check if vertex exists in the graph, if not create it
        if not self.emlg.has_vertex_collection(vertex_to_create):
            self.db.create_collection(vertex_to_create, rf)
            self.emlg.create_vertex_collection(vertex_to_create)
        else:
            logger.error("Vertex, " + vertex_to_create + " already exists!")

        return

    def remove_vertex_from_arangopipe(self, vertex_to_remove, purge=True):

        if not self.db.has_graph(self.cfg['mlgraph']['graphname']):
            self.emlg = self.db.create_graph(self.cfg['mlgraph']['graphname'])
        else:
            self.emlg = self.db.graph(self.cfg['mlgraph']['graphname'])

        #Check if vertex exists in the graph, if not create it
        if self.emlg.has_vertex_collection(vertex_to_remove):
            self.emlg.delete_vertex_collection(vertex_to_remove, purge)

            logger.info("Vertex collection " + vertex_to_remove +
                        " has been deleted!")
        else:
            logger.error("Vertex, " + vertex_to_remove + " does not exist!")

        return

    def add_edge_definition_to_arangopipe(self, edge_name, from_vertex_name,
                                          to_vertex_name):
        rf = self.cfg['arangodb'][self.mscp.DB_REPLICATION_FACTOR]

        if not self.db.has_graph(self.cfg['mlgraph']['graphname']):
            self.emlg = self.db.create_graph(self.cfg['mlgraph']['graphname'])
        else:
            self.emlg = self.db.graph(self.cfg['mlgraph']['graphname'])

        #Check if all data needed to create an edge exists, if so, create it

        if not self.emlg.has_vertex_collection(from_vertex_name):
            logger.error("Source vertex, " + from_vertex_name +\
                         " does not exist, aborting edge creation!")
            return
        elif not self.emlg.has_vertex_collection(to_vertex_name):
            logger.error("Destination vertex, " + to_vertex_name +\
                         " does not exist, aborting edge creation!")
            return

        else:
            if not self.emlg.has_edge_definition(edge_name):
                self.db.create_collection(edge_name, edge = True,\
                                          replication_factor = rf)
                self.emlg.create_edge_definition(edge_collection = edge_name,\
                                                 from_vertex_collections=[from_vertex_name],\
                                                 to_vertex_collections=[to_vertex_name] )
            else:
                logger.error("Edge, " + edge_name + " already exists!")

        return

    def remove_edge_definition_from_arangopipe(self, edge_name, purge=True):

        if not self.db.has_graph(self.cfg['mlgraph']['graphname']):
            self.emlg = self.db.create_graph(self.cfg['mlgraph']['graphname'])
        else:
            self.emlg = self.db.graph(self.cfg['mlgraph']['graphname'])

        if self.emlg.has_edge_definition(edge_name):
            self.emlg.delete_edge_definition(edge_name, purge)

        else:
            logger.error("Edge definition " + edge_name + " does not exist!")

        return

    def has_vertex(self, vertex_name):

        if not self.db.has_graph(self.cfg['mlgraph']['graphname']):
            self.emlg = self.db.create_graph(self.cfg['mlgraph']['graphname'])
        else:
            self.emlg = self.db.graph(self.cfg['mlgraph']['graphname'])

        result = self.emlg.has_vertex_collection(vertex_name)
        return result

    def has_edge(self, edge_name):

        if not self.db.has_graph(self.cfg['mlgraph']['graphname']):
            self.emlg = self.db.create_graph(self.cfg['mlgraph']['graphname'])
        else:
            self.emlg = self.db.graph(self.cfg['mlgraph']['graphname'])

        result = self.emlg.has_edge_definition(edge_name)

        return result

    def delete_all_databases(self,\
                             preserve = ['arangopipe', 'facebook_db', \
                                         'fb_node2vec_db', 'node2vecdb', '_system']):
        db_srv_host = self.cfg['arangodb'][self.mscp.DB_SERVICE_HOST]
        db_srv_port = self.cfg['arangodb'][self.mscp.DB_SERVICE_PORT]

        try:
            root_user = self.cfg['arangodb'][self.mscp.DB_ROOT_USER]
            root_user_password = self.cfg['arangodb'][
                self.mscp.DB_ROOT_USER_PASSWORD]
        except KeyError as k:
            msg = "Root credentials are unvailable, try again " + \
                     "with a new connection and credentials for root provided"
            logger.error(msg)
            logger.error("Credential information that is missing : " +
                         k.args[0])
            raise Exception("Key error associated with missing " + k.args[0])

        db_conn_protocol = self.cfg['arangodb'][self.mscp.DB_CONN_PROTOCOL]

        host_connection = db_conn_protocol + "://" + \
                db_srv_host + ":" + str(db_srv_port)
        if not root_user and not root_user_password:
            msg = "You will need to provide root credentials while connecting to perform" + \
                    " deletes of databases ! Please try again after doing so."
            logger.info(msg)
            return

        client = ArangoClient(hosts= host_connection,\
                              http_client=CustomHTTPClient(username=root_user,\
                                                           password = root_user_password))
        if not '_system' in preserve:
            preserve.append('_system')


        sys_db = client.db('_system',\
                                   username = root_user,\
                                   password = root_user_password)

        try:

            all_db = sys_db.databases()
            print("There were " + str(len(all_db) - 4) + " databases!")

            for the_db in all_db:
                if not the_db in preserve:
                    sys_db.delete_database(the_db)

        except DatabaseListError as err:
            logger.error(err)
            print("Error code: " + str(err.error_code) + " received !")
            print("Error Message: " + str(err.error_message))

        return

    def delete_database(self, db_to_delete):
        db_srv_host = self.cfg['arangodb'][self.mscp.DB_SERVICE_HOST]
        db_srv_port = self.cfg['arangodb'][self.mscp.DB_SERVICE_PORT]
        try:
            root_user = self.cfg['arangodb'][self.mscp.DB_ROOT_USER]
            root_user_password = self.cfg['arangodb'][
                self.mscp.DB_ROOT_USER_PASSWORD]
        except KeyError as k:
            msg = "Root credentials are unvailable, try again " + \
                     "with a new connection and credentials for root provided"
            logger.error(msg)
            logger.error("Credential information that is missing : " +
                         k.args[0])
            raise Exception("Key error associated with missing " + k.args[0])

        db_conn_protocol = self.cfg['arangodb'][self.mscp.DB_CONN_PROTOCOL]

        host_connection = db_conn_protocol + "://" + \
                db_srv_host + ":" + str(db_srv_port)
        if not root_user and not root_user_password:
            msg = "You will need to provide root credentials while connecting to perform" + \
                    " deletes of databases ! Please try again after doing so."
            logger.info(msg)
            return

        client = ArangoClient(hosts= host_connection,\
                              http_client=CustomHTTPClient(username=root_user,\
                                                           password = root_user_password))

        sys_db = client.db('_system',\
                               username = root_user,\
                               password = root_user_password)
        try:
            if sys_db.has_database(db_to_delete):
                sys_db.delete_database(db_to_delete)
            else:
                logger.error("The database, " + db_to_delete +
                             ", does not exist !")

        except DatabaseListError as err:
            logger.error(err)
            print("Error code: " + str(err.error_code) + " received !")
            print("Error Message: " + str(err.error_message))

        return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 14 09:17:50 2019

@author: Rajiv Sambasivan
"""
import os
import json
import yaml
from arangopipe.arangopipe_storage.managed_service_conn_parameters import ManagedServiceConnParam


class ArangoPipeConfig:
    def __init__(self):
        self.cfg = None
        self.mscp = ManagedServiceConnParam()

    def read_data(self):
        file_name = os.path.join(os.path.dirname(__file__),
                                 "arangopipe_config.yaml")
        with open(file_name, "r") as file_descriptor:
            cfg = yaml.load(file_descriptor, Loader=yaml.FullLoader)
        return cfg

    def set_cfg(self, new_cfg):
        self.cfg = new_cfg

    def get_cfg(self):
        if self.cfg is None:
            self.cfg = self.read_data()
        return self.cfg

    def export_cfg(self, file_path):
        try:
            with open(file_path, "w") as file_descriptor:
                cfg = yaml.dump(self.cfg, file_descriptor)
        except (IOError, EOFError) as e:
            print("Error writing config file to specified location:".format(
                e.args[-1]))

        return cfg

    def create_config(self, file_path):
        conn_config = {}
        try:
            with open(file_path, 'r') as stream:
                cfg = yaml.safe_load(stream)
            for key, value in cfg['arangodb'].items():
                conn_config[key] = value
        except (yaml.YAMLError, IOError, EOFError) as exc:
            print(
                "Error reading the config file from the path specified".format(
                    exc.args[-1]))

        return conn_config

    def dump_data(self):
        file_name = os.path.join(os.path.dirname(__file__),
                                 "arangopipe_config.yaml")
        with open(file_name, "w") as file_descriptor:
            cfg = yaml.dump(self.cfg, file_descriptor)
        return cfg

    def create_connection_config(self, conn_params):
        self.cfg = {
            'arangodb': {},
            'mlgraph': {
                'graphname': 'enterprise_ml_graph'
            }
        }
        for key, value in conn_params.items():
            self.cfg['arangodb'][key] = value
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Apr 13 08:35:58 2019

@author: Rajiv Sambasivan
"""

from arango import ArangoClient, AQLQueryExecuteError
import logging
from arangopipe.arangopipe_storage.arangopipe_config import ArangoPipeConfig
from arangopipe.arangopipe_storage.custom_http_client import CustomHTTPClient
from arangopipe.arangopipe_storage.managed_service_conn_parameters import ManagedServiceConnParam

# create logger with 'spam_application'
logger = logging.getLogger('arangopipe_logger')
logger.setLevel(logging.DEBUG)
# create file handler which logs even debug messages
fh = logging.FileHandler('arangopipe.log')
fh.setLevel(logging.DEBUG)
# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.ERROR)
# create formatter and add it to the handlers
formatter = logging.Formatter(
    '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
ch.setFormatter(formatter)
# add the handlers to the logger
logger.addHandler(fh)
logger.addHandler(ch)


class ArangoPipe:
    """ An instance of ArangoPipe is meant to be used to log a run of a pipeline execution. To use it:
        (1) Create a ArangoPipe object
        (2) Register your dataset with ArangoPipe
        (3) Register your featureset with ArangoPipe
        (4) Register you model with ArangoPipe
"""
    def __init__(self, config):
        self.cfg = config.get_cfg()
        self.emlg = None
        self.db = None
        self.mscp = ManagedServiceConnParam()
        self.init_graph()
        self.heart_beat()

    def heart_beat(self):
        try:
            self.lookup_dataset("heart beat check")
        except AQLQueryExecuteError as e:
            print("WARNING : " + str(e))
            logger.error(
                "Your database was perhaps deleted, try a new connection")
            #logger.error("Error: " + str(e))
            raise Exception("Your connection is stale, try a new connection!")

        return

    def get_config(self):
        apc = ArangoPipeConfig()
        return apc.get_cfg()

    def get_collection_from_id(self, id_str):
        sep = "/"
        tokens = id_str.split(sep)
        col_name = tokens[0]

        return col_name

    def is_valid_id_str(self, id_str):
        valid_id_str = False
        sep = "/"
        tokens = id_str.split(sep)

        valid_id_str = True if len(tokens) == 2 else False

        return valid_id_str

    def link_entities(self, src_id, dest_id):
        src_id_valid = self.is_valid_id_str(src_id)
        dest_id_valid = self.is_valid_id_str(dest_id)

        if not src_id_valid:
            logger.error("The source node key does appear to be valid")
            return
        if not dest_id_valid:
            logger.error("The destination key does not appear to be valid")
            return

        dest_entity_type = self.get_collection_from_id(dest_id)
        src_entity_type = self.get_collection_from_id(src_id)
        related_key = "related_" + dest_entity_type
        concat_key = 'doc.' + related_key
        aql_str = 'FOR doc in %s FILTER doc._id == @value UPDATE doc WITH {\
  %s: CONCAT_SEPARATOR(",", %s, @dest_entity) } IN %s' % (
            src_entity_type, related_key, concat_key, src_entity_type)

        cursor = self.db.aql.execute(aql_str,
                                     bind_vars={
                                         'value': src_id,
                                         'dest_entity': dest_id
                                     })

        return

    def lookup_entity_by_id(self, entity_id):
        entity_col = self.get_collection_from_id(entity_id)
        aql = 'FOR doc in %s FILTER doc._id == @value RETURN doc' % (
            entity_col)
        # Execute the query
        cursor = self.db.aql.execute(aql, bind_vars={'value': entity_id})
        asset_keys = [doc for doc in cursor]

        asset_info = None
        if len(asset_keys) == 0:
            logger.info("The asset by name: " + asset_name +\
                         " was not found in Arangopipe!")
        else:
            asset_info = asset_keys[0]

        return asset_info

    def lookup_entity(self, asset_name, asset_type):
        aql = 'FOR doc IN %s FILTER doc.name == @value RETURN doc' % (
            asset_type)
        # Execute the query
        cursor = self.db.aql.execute(aql, bind_vars={'value': asset_name})
        asset_keys = [doc for doc in cursor]

        asset_info = None
        if len(asset_keys) == 0:
            logger.info("The asset by name: " + asset_name +\
                         " was not found in Arangopipe!")
        else:
            asset_info = asset_keys[0]

        return asset_info

    def lookup_dataset(self, dataset_name):
        """ Return a dataset identifier given a name. This can be used to get the dataset id that is used to log run information associated with execution of the pipeline."""

        dataset_info = self.lookup_entity(dataset_name, 'datasets')

        return dataset_info

    def lookup_featureset(self, feature_set_name):
        """ Return a featureset identifier given a name. This can be used to get the featureset id that is used to log run information associated with execution of the pipeline."""

        featureset_info = self.lookup_entity(feature_set_name, 'featuresets')

        return featureset_info

    def lookup_model(self, model_name):
        """ Return a model identifier given a name. This can be used to get the model id that is used to log run information associated with execution of the pipeline."""

        model_info = self.lookup_entity(model_name, 'models')

        return model_info

    def lookup_modelparams(self, tag_value):
        """ Return a model parameter result given a tag."""

        # Execute the query
        cursor = self.db.aql.execute('WITH modelparams\
                                     FOR r IN run\
                    FILTER r.tag == @value \
                        FOR mp IN 1..1 OUTBOUND r run_modelparams\
                            RETURN mp ',
                                     bind_vars={'value': tag_value})
        mp_info = None
        mp_keys = [doc for doc in cursor]
        if len(mp_keys) == 0:
            logger.info("The model params for tag: " + tag_value +\
                         " was not found in Arangopipe!")
        else:
            mp_info = mp_keys[0]
        return mp_info

    def lookup_modelperf(self, tag_value):
        """ Return a model dev performance given a tag."""

        # Execute the query
        cursor = self.db.aql.execute('WITH devperf\
                                     FOR r IN run\
                    FILTER r.tag == @value \
                        FOR dp IN 1..1 OUTBOUND r run_devperf\
                            RETURN dp ',
                                     bind_vars={'value': tag_value})
        mperf_info = None
        mperf_keys = [doc for doc in cursor]
        if len(mperf_keys) == 0:
            logger.info("The model performance for tag: " + tag_value +\
                         " was not found in Arangopipe!")
        else:
            mperf_info = mperf_keys[0]

        return mperf_info

    def init_graph(self):
        """ Initialize a graph when an instance of ArangoPipe is provisioned. """
        db_serv_host = self.cfg['arangodb']['DB_service_host']
        db_serv_port = self.cfg['arangodb']['DB_service_port']
        db_name = self.cfg['arangodb']['dbName']
        db_user_name = self.cfg['arangodb']['username']
        db_passwd = self.cfg['arangodb']['password']
        db_conn_protocol = self.cfg['arangodb'][self.mscp.DB_CONN_PROTOCOL]

        host_conn_str =  db_conn_protocol +  "://" + \
                        db_serv_host + ":" + str(db_serv_port)
        client = ArangoClient(hosts= host_conn_str,\
                              http_client=CustomHTTPClient(username = db_user_name,\
                                                           password = db_passwd))

        self.db = client.db(name= db_name, \
                           username=db_user_name,\
                           password=db_passwd)

        self.emlg = self.db.graph(self.cfg['mlgraph']['graphname'])

        return




    def register_model(self, mi, user_id = "authorized_user",\
                       project = "Wine-Quality-Regression-Modelling"):
        """ Register a model. The operation requires specifying a user id. If the user id is permitted to register a model, then the registration proceeds, otherwise an unauthorized operation is indicated. """

        model_name = mi["name"]
        try:
            existing_model = self.lookup_model(model_name)
        except AQLQueryExecuteError as e:
            msg = "The model name %s is not taken" % (model_name)
            logger.info(msg)
        if existing_model is not None:
            msg = "It looks like the model name %s is already taken, try another name" % (
                model_name)
            logger.error(msg)
            return None
        models = self.emlg.vertex_collection("models")
        model_reg = models.insert(mi)

        # Execute the query
        cursor = self.db.aql.execute(
            'FOR doc IN project FILTER doc.name == @value RETURN doc',
            bind_vars={'value': project})
        project_keys = [doc for doc in cursor]
        the_project_info = project_keys[0]

        project_model_edge = self.emlg.edge_collection("project_models")
        project_model_key = the_project_info["_key"] + "-" + model_reg["_key"]

        a_project_model_edge = {"_key": project_model_key,\
                        "_from": "project/" + the_project_info["_key"],\
                        "_to": "models/" + model_reg["_key"]}
        pm_reg = project_model_edge.insert(a_project_model_edge)
        logger.info("Recording project model link " + str(pm_reg))

        return model_reg

    def register_dataset(self, ds_info, user_id="authorized_user"):
        """ Register a dataset. The operation requires specifying a user id. If the user id is permitted to register a dataset, then the registration proceeds, otherwise an unauthorized operation is indicated. """

        ds_name = ds_info["name"]
        try:
            existing_ds = self.lookup_dataset(ds_name)
        except AQLQueryExecuteError as e:
            msg = "The dataset name %s is not taken" % (ds_name)
            logger.info(msg)
        if existing_ds is not None:
            msg = "It looks like the dataset name %s is already taken, try another name" % (
                ds_name)
            logger.error(msg)
            return None
        ds = self.emlg.vertex_collection("datasets")
        ds_reg = ds.insert(ds_info)
        logger.info("Recording dataset dataset link " + str(ds_reg))

        return ds_reg


    def register_featureset(self, fs_info, dataset_id, \
                            user_id = "authorized_user"):
        """ Register a featureset. ManagedServiceConnParamThe operation requires specifying a user id. If the user id is permitted to register a featureset, then the registration proceeds, otherwise an unauthorized operation is indicated. """
        fs_name = fs_info["name"]
        try:
            existing_fs = self.lookup_featureset(fs_name)
        except AQLQueryExecuteError as e:
            msg = "The featureset name %s is not taken" % (fs_name)
            logger.info(msg)
        if existing_fs is not None:
            msg = "It looks like the featureset name %s is already taken, try another name" % (
                fs_name)
            logger.error(msg)
            return None

        fs = self.emlg.vertex_collection("featuresets")
        fs_reg = fs.insert(fs_info)
        logger.info("Recording featureset " + str(fs_reg))
        featureset_dataset_edge = self.emlg.edge_collection(
            "featureset_dataset")
        featureset_dataset_key = fs_reg["_key"] + "-" + dataset_id

        a_featureset_dataset_edge = {"_key": featureset_dataset_key,\
                        "_from": "featuresets/" + fs_reg["_key"],\
                        "_to": "datasets/" + dataset_id}
        fsds_reg = featureset_dataset_edge.insert(a_featureset_dataset_edge)
        logger.info("Recording featureset dataset link " + str(fsds_reg))

        return fs_reg

    def log_run(self, ri):
        """ Log a run. Logging a run requires specifying a dataset, featureset and a model against which this run is recored. A run records model parameters and model performance. The run object is probably most useful for the analysis of model performance with respect to a featureset, model hyper-parameters and a dataset."""

        rrid = ri["run_id"]
        mp = ri["model-params"]
        mp["_key"] = mp["run_id"]
        mperf = ri["model-perf"]
        mperf["_key"] = mperf["run_id"]
        model_key = ri["model"]

        run = self.emlg.vertex_collection("run")
        run_info = {"_key": rrid, "timestamp": mperf["timestamp"]}
        if "deployment_tag" in ri:
            run_info["deployment_tag"] = ri["deployment_tag"]
        if "tag" in ri:
            run_info["tag"] = ri["tag"]
        logger.info("Run info " + str(run_info))
        run_reg = run.insert(run_info)
        logger.info("Recording run " + str(run_reg))

        run_model_key = run_reg["_key"] + "-" + model_key
        a_run_model_edge = {"_key": run_model_key,\
                            "_from" :"models/" + model_key, \
                          "_to": "run/" + rrid}
        run_model_edge = self.emlg.edge_collection("run_models")
        rme_reg = run_model_edge.insert(a_run_model_edge)
        logger.info("Recording model run link " + str(rme_reg))

        model_param = self.emlg.vertex_collection("modelparams")
        mp_reg = model_param.insert(mp)
        logger.info("Recording model params " + str(mp_reg))

        run_fs_edge = self.emlg.edge_collection("run_featuresets")
        run_fs_key = rrid + "-" + ri["featureset"]

        a_edge_run_fs = {"_key": run_fs_key,\
                        "_from": "run/" + rrid,\
                        "_to": "featuresets/" +  ri["featureset"]}

        rfse_reg = run_fs_edge.insert(a_edge_run_fs)
        logger.info("Recording run featureset link " + str(rfse_reg))

        run_mp_edge = self.emlg.edge_collection("run_modelparams")
        run_mp_key = rrid + "-" + mp["run_id"]

        a_run_mp_edge = {"_key": run_mp_key,\
                        "_from": "run/" + rrid,\
                        "_to": "modelparams/" + mp_reg["_key"]}

        rmp_reg = run_mp_edge.insert(a_run_mp_edge)
        logger.info("Recording run model params " + str(rmp_reg))

        model_perf = self.emlg.vertex_collection("devperf")
        dp_reg = model_perf.insert(mperf)
        logger.info("Recording model dev performance  " + str(dp_reg))

        run_devperf_edge = self.emlg.edge_collection("run_devperf")
        run_devperf_key = rrid + "-" + dp_reg["_key"]

        a_run_devperfedge = {"_key": run_devperf_key,\
                        "_from": "run/" + rrid,\
                        "_to": "devperf/" + dp_reg["_key"]}
        rdp_reg = run_devperf_edge.insert(a_run_devperfedge)
        logger.info("Recording run dev perf link " + str(rdp_reg))

        run_dataset_edge = self.emlg.edge_collection("run_datasets")
        run_dataset_key = rrid + "-" + ri["dataset"]

        a_run_dataset_edge = {"_key": run_dataset_key,\
                        "_from": "run/" + rrid,\
                        "_to": "datasets/" + ri["dataset"]}
        rds_reg = run_dataset_edge.insert(a_run_dataset_edge)
        logger.info("Recording run dev perf link " + str(rds_reg))

        return

    def log_serving_perf(self, sp, dep_tag, userid="authorized user"):
        """ Log serving performance against a deployed model. The user making the request needs to be authorized to log this performance update. A serving performance vertex is created and is linked with its deployment vertex"""
        servingperf = self.emlg.vertex_collection("servingperf")
        sp_reg = servingperf.insert(sp)

        # Execute the query
        cursor = self.db.aql.execute(
            'FOR doc IN deployment FILTER doc.tag == @value RETURN doc',
            bind_vars={'value': dep_tag})
        dep_docs = [doc for doc in cursor]
        the_dep_doc = dep_docs[0]
        # Link the service performance record with the deployment record
        dep_servingperf_edge = self.emlg.edge_collection(
            "deployment_servingperf")
        dep_servingperf_key = the_dep_doc["_key"] + "-" + sp_reg["_key"]
        the_dep_servingperf_edge = { "_key": dep_servingperf_key,\
                                "_from": the_dep_doc["_id"],\
                                "_to": sp_reg["_id"]}

        dep_servingperf_reg = dep_servingperf_edge.insert(
            the_dep_servingperf_edge)
        return dep_servingperf_reg

    def insert_into_vertex_type(self, vertex_type_name, document):
        vertex_info = None
        if self.emlg.has_vertex_collection(vertex_type_name):
            vc = self.emlg.vertex_collection(vertex_type_name)
            vertex_info = vc.insert(document)
        else:
            logger.error("Vertex, " + vertex_type_name +
                         " does not exist in Arangopipe!")

        return vertex_info

    def insert_into_edge_type(self,
                              edge_name,
                              from_vdoc,
                              to_vdoc,
                              document=None):
        edge_info = None
        if self.emlg.has_edge_collection(edge_name):
            try:
                ec = self.emlg.edge_collection(edge_name)
                edge_key = from_vdoc['_key'] + "-" + to_vdoc['_key']
                if document is not None:
                    document["_from"] = from_vdoc['_id']
                    document["_to"] = to_vdoc['_id']
                    document["_key"] = edge_key
                    edge_info = ec.insert(document)
                else:
                    document = { "_from": from_vdoc['_id'],\
                                "_to": to_vdoc['_id'],\
                                "_key": edge_key }
                    edge_info = ec.insert(document)
            except Exception as e:
                logger.error(e)
        else:
            logger.error("Edge, " + edge_name + " does not exist!")

        return edge_info



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec  5 09:05:19 2019

@author: Rajiv Sambasivan
"""


class ManagedServiceConnParam:
    @property
    def DB_SERVICE_HOST(self):
        return "DB_service_host"

    @property
    def DB_SERVICE_END_POINT(self):
        return "DB_end_point"

    @property
    def DB_SERVICE_NAME(self):
        return "DB_service_name"

    @property
    def DB_SERVICE_PORT(self):
        return "DB_service_port"

    @property
    def DB_NAME(self):
        return "dbName"

    @property
    def DB_REPLICATION_FACTOR(self):
        return "arangodb_replication_factor"

    @property
    def DB_USER_NAME(self):
        return "username"

    @property
    def DB_PASSWORD(self):
        return "password"

    @property
    def DB_ROOT_USER(self):
        return "root_user"

    @property
    def DB_ROOT_USER_PASSWORD(self):
        return "root_user_password"

    @property
    def DB_CONN_PROTOCOL(self):
        return "conn_protocol"

    @property
    def DB_NOTIFICATION_EMAIL(self):
        return "email"
